#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Gterminal pluggable gobject-io readline magic terminal like interface.

This is a generic terminal frontend for akin to the mysql> or bash$ interfaces
which uses gobject and select() for it's input/output. It magically rewrites
the screen to dynamically display status messages and other important
information. The generic classes provided make it suitable for use with
existing projects. It supports bash-style history and cursor movements.
"""
# Copyright (C) 2008-2010  James Shubin, McGill University
# Written for McGill University by James Shubin <purpleidea@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# NOTES:
# * Colour Schemes
# As a rough guideline, users will notice that text in green is used for normal
# situations in which objects are enhanced by being coloured. Blue is used when
# the user should focus their attention to a particular event or happening. Red
# is used when there is some sort of error condition or bad thing that happens.

# TODO: which variables/functions should be _local and which should be __private ?
import os
import sys
import math
import atexit
import termios
import tty
import gobject
from jhelp import gobject_once
import string
import curses.ascii
import signal
import errno
import readline		# TODO: add coloured tab completion, etc...
			# TODO: on bell flash the `gterminal>` part if possible
# TODO: should this be replaced with glib.get_* functions?
# http://www.pygtk.org/docs/pygobject/glib-functions.html#function-glib--get-user-config-dir
import xdg.BaseDirectory
import logging
from jhelp import logginghelp
from jhelp import tcrecipe
import gcmd		# basic command class for gterminal

_ = lambda x: x			# add fake gettext function until i fix up i18n
__all__ = ['gterminal']
DEBUG = True

class gterminal:
	"""Class to encapsulate the magic readline functionality."""

	def __init__(self, name, cmdobj=None, logger=None,
		ps1='%s> ' % os.path.splitext(os.path.basename(__file__))[0]):
		"""Be able to build and customize your own gterminal object."""
		# MISC ########################################################
		self.name = name
		self.ps1 = ps1
		self.good = string.digits + string.letters + ' '
		self.good += string.punctuation		# we need these too
		self.mainloop = gobject.MainLoop()	# include a mainloop

		# save initial terminal settings
		self.__termios = termios.tcgetattr(sys.stdin)
		self.__restored = False

		self.__messages = {}			# dictionary of msgs
		self.__rmsg = None			# handle for resize msg

		# LOGGING #####################################################
		# NOTE: while for all our purposes it's probably fine to pass a
		# Logger to this class, it could be beneficial to pass it a log
		# object that we can call our own get_log() on. this way we can
		# potentially allocate additional loggers for children objects.
		if isinstance(logger, logging.Logger): self.log = logger
		else:	# if we aren't provided with a logger obj then make one
			obj = logginghelp.logginghelp(
				name=self.name, wordymode=True,
				stderrlog=False, mylogpath=[], mykerning=1
			)
			self.log = obj.get_log()	# main logger

		# XDG #########################################################
		# uncomment below two lines for adding a config directory
		#self.__conf = xdg.BaseDirectory.load_first_config(self.name) or
		#xdg.BaseDirectory.save_config_path(self.name)
		self.__data = (
			list(xdg.BaseDirectory.load_data_paths(self.name)) or
			[xdg.BaseDirectory.save_data_path(self.name)]
		)[0]	# take the first element in the resultant list

		# HISTORY #####################################################
		histfile = os.path.join(self.__data, 'history')
		try: readline.read_history_file(histfile)
		except IOError: pass
		atexit.register(readline.write_history_file, histfile)

		# DISPLAY #####################################################
		self.tc = tcrecipe.TerminalController()
		self.__block = tcrecipe.Block(self.tc)
		self.__buffer = ''		# text to display on screen
		self.__latest = ''		# text buffer for last item
		self.__xpos = 0			# position of the cursor
		self.__ypos = self.__ypos_index_latest()	# cmd hx pos

		# CMD #########################################################
		# cmd class must provide run() and can provide set_gterminal() 
		if cmdobj is None:
			cmdobj = gcmd.base_gcmd()
		# check if class has minimum required methods
		elif not hasattr(cmdobj, 'run'):
			msg = _('Command object replaced with stock object.')
			self.log.warn(msg)
			cmdobj = gcmd.base_gcmd()

		self.__cmdobj = cmdobj
		# pass a ref of the main gterminal for sending messages, etc...
		if hasattr(cmdobj, 'set_gterminal'):
			self.__cmdobj.set_gterminal(self)

		# GO GO GO... #################################################
		tty.setcbreak(sys.stdin, termios.TCSANOW)	# setup terminal
		gobject.io_add_watch(sys.stdin, gobject.IO_IN, self.__io_cb)
		#gobject.io_add_watch(sys.stdin, gobject.IO_HUP, self.__io_cb)
		self.__block.update(self.__prompt())	# display initial prompt
		atexit.register(self.restore)		# register exit handler!
		signal.signal(signal.SIGINT, self.__ctrl_c)	# handle CTRL-C
		signal.signal(signal.SIGWINCH, self.__sigwinch)	# handle resize

		# OLD/TESTING #################################################
		# capture stdin
		#import fcntl
		#flags = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
		#fcntl.fcntl(sys.stdin, fcntl.F_SETFL, flags | os.O_NONBLOCK)


	# MISC ################################################################
	def restore(self):
		"""Restore original terminal settings."""
		if not self.__restored:
			termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.__termios)
			self.__restored = True	# only let this be run once


	def run(self):
		"""Get gterminal going..."""
		# when mainloop.quit() is called, the blocking run() call below
		# unblocks and gives way to the clear call. then this all ends.
		self.mainloop.run()
		self.__block.clear()	# get rid of erroneous extra prompt


	def quit(self):
		"""Cause gterminal to exit."""
		self.mainloop.quit()


	# SIGNALS #############################################################
	def __sigwinch(self, signum, frame):
		"""This signal handler called on terminal resize events."""
		self.tc.refresh()				# refresh vals
		self.__block.settc(self.tc)			# also in case
		msg = _('screen resized to: %d x %d')
		msg = msg % (self.tc.COLS, self.tc.LINES)
		self.__rmsg = self.msg(msg, lifetime=1, name=self.__rmsg)


	def __ctrl_c(self, signum, frame):
		"""Signal handler for CTRL-C"""
		self.__safe('^C')		# display a ^C
		self.__buffer = ''		# reset the buffer
		self.__update()			# refresh the prompt


	# MESSAGING ###########################################################
	def __message_id(self):
		"""Returns an available message id for use in a new message."""
		tmp = [0]		# list of id's that are not available
		m = self.__messages	# dictionary of messages
		while True:
			i = max([x for x in m.keys()+tmp if type(x) is int])+1
			if i in m:
				# not available
				tmp.append(i)
			else:
				return i


	def msg(self, message, lifetime=15, name=None):
		"""Add a notification message above the prompt. If lifetime is
		None, then it never expires. This returns a message handle."""
		# when will it expire... (and dissapear)
		now = gobject.get_current_time()
		if lifetime is None: expiry = None
		else: expiry = int(math.floor(now + lifetime))

		if name is None:
			name = self.__message_id()
		# if message already exists, don't modify the now field.
		if name in self.__messages: now = self.__messages[name][2]

		self.__messages[name] = ((message, expiry, now ))

		gobject.timeout_add_seconds(0, self.__update)	# update now
		# and schedule an update for the expected expiry time
		if lifetime is not None:
			gobject.timeout_add_seconds(lifetime, self.__update)
		return name


	def msg_clear(self, handle):
		"""Schedules the immediate clear of a particular message."""
		if handle in self.__messages:
			del self.__messages[handle]
			gobject.timeout_add_seconds(0, self.__update)


	# DISPLAY #############################################################
	def __prompt(self, buf=''):
		"""The prompt."""
		tmp = '${GREEN}%s${NORMAL}%s' % (self.ps1, buf)
		return self.tc.render(tmp)


	def __update(self):
		"""Update / refresh the prompt. Does message scheduling too."""
		# filter out old messages by dictionary comprehension
		self.__messages = dict([ (key,(msg,exp,now)) for
		key,(msg,exp,now) in self.__messages.items() if
		exp > gobject.get_current_time() or exp is None ])
		lines = []
		# to display sensibly, sort by time added (the `now' field)
		# we want to see the *newest* items at the bottom near the >
		items = self.__messages.items()
		# [1] is the value, [2] is the now field (the third one)
		items.sort(key=lambda x: x[1][2])
		for (name, (msg, exp, now)) in items:
			line = '${BLUE}@${NORMAL} %s' % msg
			lines.append(self.tc.render(line))

		lines.append(self.__prompt(self.__buffer))
		self.__block.clear()
		self.__block.update(lines)

		# lastly, position the cursor where we expect it to be
		for i in range(len(self.__buffer) - self.__xpos):
			sys.stdout.write(self.tc.LEFT)
		sys.stdout.flush()
		return False


	def __safe(self, text=None, force=False):
		"""Adds protector lines to avoid any messages squashing output.
		The force option should be used if have previously printed text
		manually, and you require the necessary padding to be added."""
		# this should only be called once per __update() call. multiple
		# calls interweaved with __update() cause it to seem as if many
		# commands have passed. not generally recommended for use ever.
		def pad():
			"""Simple padding function required to protect text."""
			for i in self.__messages: print ''

		if text is not None:
			print text
			# this adds newlines so that any above
			# run text doesn't get munched over by
			# the messages that get printed. print
			# out one line for every message text.
			pad()
		elif force: pad()


	# IO ##################################################################
	def __getch(self, source):
		"""This should be called only when you either expect some data
		from the source, or are happy to have it block until some data
		arrives. It returns an array of numeric codes to process, but
		usually you'll just get one element. We only return the ord
		because for some possibilities (eg: >=256, the chr function
		can't handle the result and) we don't want to print an up arrow
		character anyways (whatever it might be). This func. uses yield
		magic, and the only way key presses can cause gobject to really
		block and wait for release, is by repeated <esc> pressing."""
		# NOTE: this was a *very* tricky/hard function to write. don't
		# change *anything* unless you know exactly what you're doing!
		SPECIAL_ORD = 91
		SPECIAL_MAP = {
			65: curses.KEY_UP, 66: curses.KEY_DOWN,
			67: curses.KEY_RIGHT, 68: curses.KEY_LEFT,
			51: curses.KEY_DC
		}
		sig = False	# previously named `esc'. now named sig.
		special = False
		munch = False
		while True:
			# *** capturing section ***
			if sig:
				signal.signal(signal.SIGALRM, lambda x,y: None)
				# sends '\0' on alarm
				signal.set_wakeup_fd(source.fileno())
				# FIXME: use ualarm to make time less than 1 sec
				signal.alarm(1)
			try:
				# it is very important to first set an initial
				# value for c, because if an error occurs when
				# trying to read, then the c will retain a bad
				# value (the last one it succesfully got), and
				# this looks like repeated keypresses. usually
				# it will look like repeated presses to escape
				c = '\0'	# this line is very important!
				c = source.read(1)
			except IOError, e:
				if errno.EINTR == e.errno: pass
				else: raise e
			finally:
				if sig: signal.alarm(0)	# disable the alarm
			o = ord(c)

			# *** processing section ***
			if munch:	# 1 char has been eaten. we can quit now
				assert o == munch, 'expected: chr(%d), got: chr(%d)' % (munch, o)
				break

			elif o == 0 and sig:
				# esc: last or singly pressed, yield and return
				yield curses.ascii.ESC
				break		# we're done, go return

			elif o == curses.ascii.ESC:
				special = False	# avoid sequence: <esc>[<esc>A
				# if `sig' was *already* true before we set it
				# below, then we've received one escape press,
				# followed by an escape, or arrow key press,
				# which hasn't been determined yet. (so loop)
				if sig:
					# sig: when pressed in rapid succession
					# the magic: let the caller get the esc
					# value now, so that we don't pause the
					# execution! this is the magic of yield
					yield curses.ascii.ESC
				sig = True

			# magic char after an esc
			elif sig and o == SPECIAL_ORD:
				# we have a special key?
				special = True
				sig = False

			elif special:
				if o in SPECIAL_MAP.keys():
					yield SPECIAL_MAP[o]
					# there seems to be an extra `126'
					# that hangs around after a delete
					# press, so gobble it out of the q
					if SPECIAL_MAP[o] == curses.KEY_DC:
						munch = 126
						sig = True	# to add timeout
					else:
						break
				else:
					if DEBUG:
						self.msg('DEBUG: special: chr(%d) not handled' % o)
					yield o
					break

			# if we get a null and sig is False, then the alarm
			# probably fired prematurely. it's probably fine to
			# return the null character and let the caller deal
			#elif o == 0:
			#	yield o
			#	break
			else:	# regular character
				yield o
				break
		return


	def __io_in(self, source):
		"""The io_in processor. Broken up into this function to keep my
		code flat, and less nested. This should be called by __io_cb"""
		# NOTE: the ordering of the big if statement below is important!

		# loop in case there are multiple simultaneous presses
		for o in self.__getch(source):	# receives yields

			# avoid printing characters that we don't want to print
			if o not in [curses.ascii.ESC, curses.ascii.EOT]:
				try:
					temp = chr(o)
					# print out the character
					sys.stdout.write(temp)
					# flush is required to show it
					sys.stdout.flush()
				except ValueError:
					pass

			# NOTE: the ordering of this if statement is important!
			if o == curses.ascii.NL:	# 10: <enter> (run)
				if len(self.__buffer) > 0:
					readline.add_history(self.__buffer)
					# run the cmd and print the output
					out = self.__cmdobj.run(self.__buffer)
					# don't print nothing or empty strings!
					if isinstance(out, str) and out != '':
						print out

					# now, update the pos to be most recent
					self.__xpos = 0
					self.__ypos = self.__ypos_index_latest()
					self.__buffer = ''	# now erase it
				else:
					pass

				self.__safe(force=True)		# protect

			elif o == curses.ascii.TAB:	# 9: <tab> (complete)
				# TODO: implement some sort of tab completion
				pass

			elif o == curses.ascii.FF:	# 12: ^L (clear)
				sys.stdout.write(self.tc.CLEAR_SCREEN)

			elif o == curses.ascii.EOT:	# 4: ^D (end of text)
				# bash only accepts a ^D if the buffer is empty
				if self.__buffer == '':
					self.__safe('exit')
					self.quit()

			elif o == curses.ascii.ESC:	# 27: <esc>
				#self.__safe('ESC')
				# nothing better to do for now
				sys.stdout.write(self.tc.BELL)

			elif o == curses.KEY_LEFT:	# cursor: <prev>
				if not self.__cursor_prev():
					# beep if nowhere to move cursor
					sys.stdout.write(self.tc.BELL)

			elif o == curses.KEY_RIGHT:	# cursor: <next>
				if not self.__cursor_next():
					# beep if nowhere to move cursor
					sys.stdout.write(self.tc.BELL)

			elif o == curses.KEY_UP:	# hx: <prev>
				if not self.__history_prev():
					# beep if no past history to browse
					sys.stdout.write(self.tc.BELL)

			elif o == curses.KEY_DOWN:	# hx: <next>
				if not self.__history_next():
					# beep if no forward history to browse
					sys.stdout.write(self.tc.BELL)

			elif o == curses.KEY_DC:	# <delete>
				if not self.__cursor_del():
					# beep if we're out of buffer to remove
					sys.stdout.write(self.tc.BELL)

			# FIXME: which is it? 127 or 8 or both?
			elif (o == curses.ascii.DEL or	# 127: <backspace>
			o == curses.ascii.BS):		# 8: (on sam's computer)
				if not self.__cursor_bs():
					# beep if we're out of buffer to remove
					sys.stdout.write(self.tc.BELL)

			else:
				if not self.__cursor_add(o):
					# beep on `bad' chars
					if DEBUG: self.msg('DEBUG: unknown: chr(%d) not handled' % o)
					sys.stdout.write(self.tc.BELL)

			# XXX: maybe add a self.__safe() here, and allow people
			# to print whatever they like above. either that or add
			# all text to a common buffer and then print it all here
			self.__update()			# refresh the prompt

		return True


	def __io_cb(self, source, cb_condition):
		"""I/O handler."""
		if cb_condition == gobject.IO_IN:
			return self.__io_in(source)

		elif cb_condition == gobject.IO_HUP:
			# FIXME: does this ever happen ?
			# what should we do if it does ?
			self.log.critical(_('IO_HUP received.'))
			return False
		else:
			msg = _('Unexpected callback: %d') % cb_condition
			self.log.error(msg)
			return False


	# HISTORY #############################################################
	def history(self, last=None):
		"""Output the history to a nicely readable string format. You
		can specify a `last' integer to display only the last few."""
		l = readline.get_current_history_length()
		# add correct padding for history display
		pattern = '%' + str(len(str(l))) + 'd  %s'	# two spaces b/w
		output = []
		start = 1
		if last is not None: start = l+1-last
		for i in range(max(start, 1), l+1):
			txt = readline.get_history_item(i)
			output.append(pattern % (i, txt))

		return '\n'.join(output)


	def __ypos_index_latest(self):
		"""Returns magic index of latest element."""
		return readline.get_current_history_length() + 1


	def __is_latest(self):
		"""Returns whether ypos is referring to the last element."""
		return self.__ypos == self.__ypos_index_latest()


	def __history_replace(self):
		"""Puts all the magic of adding and removing characters from
		readline databases into one function that is easy to call."""
		# update any items except the latest one
		if not self.__is_latest():
			readline.replace_history_item(self.__ypos-1, self.__buffer)


	def __history_prev(self):
		"""Command to cause previous history to be displayed."""
		if self.__ypos > 1:	# readline is 1 indexed
			# save latest
			if self.__is_latest(): self.__latest = self.__buffer
			self.__ypos -= 1	# update ypos
			self.__buffer = readline.get_history_item(self.__ypos)
			self.__xpos = len(self.__buffer)	# update cursor
			return True
		else: return False	# no past history to browse


	def __history_next(self):
		"""Command to cause succeeding history to be displayed."""
		if self.__ypos < self.__ypos_index_latest():
			self.__ypos += 1
			# load latest
			if self.__is_latest(): self.__buffer = self.__latest
			else: self.__buffer = readline.get_history_item(self.__ypos)
			self.__xpos = len(self.__buffer)	# update cursor
			return True
		else: return False	# no forward history to browse


	# CURSOR ##############################################################
	def __cursor_prev(self):
		"""Command to cause the cursor to move to the left."""
		if self.__xpos > 0:
			self.__xpos -= 1	# update xpos
			return True
		else: return False


	def __cursor_next(self):
		"""Command to cause the cursor to move to the right."""
		if self.__xpos < len(self.__buffer):
			self.__xpos += 1	# update xpos
			return True
		else: return False


	def __cursor_del(self):
		"""Command to delete a char from before the current xpos."""
		# this condition check if we can do a <delete>
		if self.__xpos < len(self.__buffer):
			self.__buffer = self.__buffer[0:self.__xpos] + self.__buffer[self.__xpos+1:]
			# xpos *doesn't* get updated for this one.
			self.__history_replace()	# update
			return True
		else: return False


	def __cursor_bs(self):
		"""Command to delete a char from the current xpos."""
		# NOTE: we used to check the size of buffer to
		# ensure there actually were characters to del
		# but since we can assume buffer size via xpos
		# it makes more sense to just check that which
		# also lets you know if you can delete at left
		if self.__xpos > 0:
			# del the char from the buffer at xpos
			self.__buffer = self.__buffer[0:self.__xpos-1] + self.__buffer[self.__xpos:]
			self.__xpos -= 1
			self.__history_replace()	# update
			return True
		else: return False


	def __cursor_add(self, o):
		"""Command to insert a char from the current xpos."""
		try:
			if not chr(o) in self.good: return False
		except ValueError, e:
			return False
		# add the character to the buffer at xpos
		self.__buffer = self.__buffer[0:self.__xpos] + chr(o) + self.__buffer[self.__xpos:]
		self.__xpos += 1	# increment the position
		self.__history_replace()		# update
		return True


if __name__ == '__main__':
	# example
	name = os.path.splitext(os.path.basename(__file__))[0]
	obj = gterminal(name)
	gobject.once_add_seconds(5, obj.msg, 'This is an example message.')
	obj.run()

